
{-------------------------------------------------------------------------------
 -                                                                             -
 - Simulador y simplificador rudimientario de Automatas Finitos Deterministas  -
 -                                                                             - 
 - Creado por: Alejandro Dobeson Beirao                                        -
 - Contacto: granmacco@gmail.com ; www.sandoreus.net                           -
 -                                                                             -
 - Este software se distribuye "tal cual" bajo licencia aGPL v3.0              -
 - Compruebe los terminos dicha licencia en el archivo LICENSE.TXT incluido    -
 -                                                                             -
 -------------------------------------------------------------------------------}


module AFDdata (Delta, AFD, modificaDelta, anadeDelta, anadeEstado, eliminaChar, eliminaEstado, arcosHaciaQ, hayArcosHaciaQ, inBounds, inSigma, ejecutaAFD, minimizaAutomata, ordenaSigma, sonIguales) where

import Array
import Hugs.Observe

type Delta = Array (Int, Char) Int
type AFD = ([Char], Int, Int, [Int], Delta)

-- Array de prueba: array ((1, 'a'), (2, 'a')) [((1,'a'), 1), ((2,'a'),2)]

-- afd complejo: ("ab", 9, 4, [3, 5, 8], array ((1,'a'),(9, 'b')) [((1,'a'),1),((1,'b'),1),((2,'a'),2),((2,'b'),3),((3,'a'),5),((3,'b'),5),((4,'a'),3),((4,'b'),4),((5,'a'),3),((5,'b'),3),((6,'a'),3),((6,'b'),3),((7,'a'),3),((7,'b'),3),((8,'a'),3),((8,'b'),3),((9,'a'),3),((9,'b'),3)])
-- afd complejo tras eliminar estados inalcanzables: ("ab",3,2,[1,3],array ((1,'a'),(3,'b')) [((1,'a'),3),((1,'b'),3),((2,'a'),1),((2,'b'),2),((3,'a'),1),((3,'b'),1)])
--
-- read [Char]
-- read Int
-- read Int
-- read [Int]
-- read Array

-- 
-- read AFD

--leerAFD :: String -> IO (AFD)
--leerAFD x = do cont <- readFile x
--               return (read cont)
--             let (cont1:cont2:cont3:cont4:cond5) = lines cont
--             return ((cont1, read cont2, read cont3, read cont4, leerDelta cont5))

--f :: String -> Int
--f x = read x

modificaDelta::AFD->Int->Char->Int->AFD
modificaDelta (a, b, c, d, e) qi elem qf = (a, b, c, d, (modificaDelta' e qi elem qf))
      
modificaDelta'::Delta->Int->Char->Int->Delta
modificaDelta' d qi e qf = d // [((qi, e), qf)]

ordenaSigma::[Char]->[Char]
ordenaSigma [] = []
ordenaSigma (x:xs) = (filter (<x) xs) ++ [x] ++ (filter (>=x) xs)

anadeDelta::AFD->[((Int, Char), Int)]->AFD
anadeDelta (a, b, c, d, e) assocs = (a, b, c, d, (array ((1, (head a)), (b, (last a))) assocs))

anadeEstado::AFD->Int->AFD
anadeEstado afd@([], b, c, d, e) n = ([], b+n, c, d, e)
anadeEstado afd@(a, b, c, d, e) n = anadeEstado' (a, (b+n), c, d, e) (b + 1) (assocs e)

anadeEstado'::AFD->Int->[((Int, Char),Int)]->AFD
anadeEstado' afd@(a, b, c, d, e) q ys = if (b < q) then (a, b, c, d, (array ((1, (head a)),(b, (last a))) ys))
                                                   else anadeEstado' afd (q+1) ((zip (zip (repeat q) a) (repeat q)) ++ ys)

creaArrayVacio::Delta
creaArrayVacio = array ((0, 'b'), (0, 'a')) []
                                                   
eliminaChar::AFD->Char->AFD
eliminaChar (a, b, c, d, e) ch = if ((length a) == 1) then ([], b, c, d, creaArrayVacio)
                                                      else eliminaChar' (filter (/=ch) a, b, c, d, e) ch

eliminaChar'::AFD->Char->AFD
eliminaChar' (a, b, c, d, e) ch = (a, b, c, d, array ((1, (head a)), (b, (last a))) (filtrado (assocs e)))
  where filtrado [] = []
        filtrado (x:xs) = if ((snd(fst x)) == ch) then (filtrado xs)
                                                  else x:(filtrado xs)

eliminaEstado::AFD->Int->AFD
eliminaEstado afd@(a, b, c, d, e) q = if (q < c) then eliminaEstado' (a, b, c - 1, mapea (filter (/=q) d), e) q 1 a
                                                 else eliminaEstado' (a, b, c, mapea (filter (/=q) d), e) q 1 a
  where mapea [] = []
        mapea (x:xs) = if (q < x) then (x-1):(mapea xs)
                                  else x:(mapea xs)

eliminaEstado'::AFD->Int->Int->[Char]->AFD
eliminaEstado' afd@(a, b, c, d, e) qElim qActual [] = eliminaEstado' afd qElim (qActual + 1) a
eliminaEstado' afd@(a, b, c, d, e) qElim qActual (x:xs)
  | qActual > b = (a, b - 1, c, d, (array ((1, (head a)), (b - 1, (last a))) (reverse (drop (length a) (reverse (assocs e))))))
  | qActual < qElim = if (qDestino > qElim) then eliminaEstado' (modificaDelta afd qActual x (qDestino - 1)) qElim qActual xs
                                            else eliminaEstado' afd qElim qActual xs
  | qActual > qElim = if (qDestino > qElim) then eliminaEstado' (modificaDelta afd (qActual - 1) x (qDestino - 1)) qElim qActual xs
                                            else eliminaEstado' (modificaDelta afd (qActual - 1) x qDestino) qElim qActual xs
  | otherwise = eliminaEstado' afd qElim (qActual + 1) a
  where qDestino = (e ! (qActual, x))

hayArcosHaciaQ::AFD->Int->Bool
hayArcosHaciaQ afd q = not (null (filtraArcosReflexivos (arcosHaciaQ afd q)))
  where filtraArcosReflexivos [] = []
        filtraArcosReflexivos (x:xs) = if ((fst (fst x)) == (snd x)) then filtraArcosReflexivos xs
                                                                     else x:(filtraArcosReflexivos xs)

arcosHaciaQ::AFD->Int->[((Int, Char),Int)]
arcosHaciaQ afd@(a, b, c, d, e) q = filter (\x -> (snd x) == q) (assocs e)

arcosHaciaQsinReflex::AFD->Int->[((Int, Char),Int)]
arcosHaciaQsinReflex afd q = filter (\x -> ((fst.fst) x) == (snd x)) (arcosHaciaQ afd q)

inBounds::Int->Int->Bool
inBounds a b = (b > 0) && b < (a+1)

inSigma::[Char]->[Char]->Bool
inSigma xs [] = True
inSigma xs (y:ys) = (inSigma' xs y) && (inSigma xs ys)

inSigma'::[Char]->Char->Bool
inSigma' xs x = not (null (filter(==x) xs))

ejecutaAFD::AFD->Int->String->Int
ejecutaAFD afd@(a, b, c, d, e) qi [] = qi
ejecutaAFD afd@(a, b, c, d, e) qi (x:xs) = ejecutaAFD afd (e ! (qi, x)) xs

minimizaAutomata::AFD->AFD
minimizaAutomata ([], b, c, d, e) = ([], b, c, d, e)
minimizaAutomata (a, b, 1, d, e) = (a, b, 1, d, e)
minimizaAutomata afd = eliminaEstadosNoDistinguibles (eliminaEstadosInalcanzables afd)

----------------------------------------------------------
-- Funciones relacionadas con hallar estados inalcanzables
----------------------------------------------------------

eliminaEstadosInalcanzables::AFD->AFD
eliminaEstadosInalcanzables afd = eliminaEstadosInalcanzables' afd (reverse (testInalcanzabilidad afd))

eliminaEstadosInalcanzables'::AFD->[Int]->AFD
eliminaEstadosInalcanzables' afd [] = afd
eliminaEstadosInalcanzables' afd@(a, b, c, d, e) (x:xs) = eliminaEstadosInalcanzables' (eliminaEstado afd x) xs 

ordenaLista::Ord a=>[a]->[a]
ordenaLista [] = []
ordenaLista (x:xs) = (ordenaLista (filter (<x) xs)) ++ [x] ++ (ordenaLista (filter (>x) xs))

eliminaRepetidos::Eq a=>[a]->[a]
eliminaRepetidos [] = []
eliminaRepetidos (x:xs) = x:(eliminaRepetidos (filter (/= x) xs))

testInalcanzabilidad::AFD->[Int]
testInalcanzabilidad afd@(a, b, c, d, e) = ordenaLista( eliminaRepetidos( testInalcanzabilidad' 1 b (testAlcanzabilidad afd)))

testInalcanzabilidad'::Int->Int->[Int]->[Int]
testInalcanzabilidad' qActual qFinal qAlcanzables = if (qActual <= qFinal) then if (qActual `notElem` qAlcanzables) then qActual:(testInalcanzabilidad' (qActual + 1) qFinal qAlcanzables)
                                                                                                                    else (testInalcanzabilidad' (qActual + 1) qFinal qAlcanzables)
                                                                           else []

testAlcanzabilidad::AFD->[Int]
testAlcanzabilidad afd@(a, b, c, d, e) = ordenaLista (eliminaRepetidos (testAlcanzabilidad' afd [c] []))

testAlcanzabilidad'::AFD->[Int]->[Int]->[Int]
testAlcanzabilidad' afd [] xs = []
testAlcanzabilidad' afd@(a, b, c, d, e) (q:qs) xs = if (q `notElem` xs) then qAlcanzados ++ (testAlcanzabilidad' afd (qAlcanzados ++ qs) (q:xs))
                                                                        else testAlcanzabilidad' afd qs xs
  where qAlcanzados = estadosAlcanzables afd q

estadosAlcanzables::AFD->Int->[Int]
estadosAlcanzables afd@(a, b, c, d, e) q = map (snd) transiciones
  where transiciones = filter ((== q).fst.fst) (assocs e)

-----------------------------------------------------------------------------------------
-- Funciones relacionadas con hallar la lista de estados distinguibles y no distinguibles
-----------------------------------------------------------------------------------------
  
eliminaEstadosNoDistinguibles::AFD->AFD
eliminaEstadosNoDistinguibles afd@(a, b, c, d, e) = eliminaEstadosNoDistinguibles' afd (hallaParesNoDistinguibles afd)

eliminaEstadosNoDistinguibles'::AFD->[(Int, Int)]->AFD
eliminaEstadosNoDistinguibles' afd [] = afd
eliminaEstadosNoDistinguibles' afd@(a, b, c, d, e) (x:xs) = eliminaEstadosNoDistinguibles' (eliminaEstado (cambiaArcosDeQaQ (a, b, (compruebaEstadoInicial c (fst x) (snd x)), d, e) arcosAcambiar (fst x)) (snd x)) xs
  where arcosAcambiar = map (fst) (arcosHaciaQ afd (snd x))
        compruebaEstadoInicial qIni q1 q2 = if (qIni == (max q1 q2) ) then (min q1 q2)
                                                                      else qIni

cambiaArcosDeQaQ::AFD->[(Int, Char)]->Int->AFD
cambiaArcosDeQaQ afd [] qNuevo = afd
cambiaArcosDeQaQ afd (arco:arcos) qNuevo = cambiaArcosDeQaQ (modificaDelta afd (fst arco) (snd arco) qNuevo) arcos qNuevo

hallaParesNoDistinguibles::AFD->[(Int, Int)]
hallaParesNoDistinguibles afd@(a, b, c, d, e) = filtrado paresTotales paresDistinguibles
  where paresDistinguibles = bucleMientrasCambie afd (estadosDistinguiblesTriviales afd)
        paresTotales = zipEspecial [1..b]
        zipEspecial [] = []
        zipEspecial (x:xs) = if (not (null xs)) then (zip (repeat x) [x+1..(last xs)]) ++ (zipEspecial xs)
                                              else []
        filtrado [] ys = []
        filtrado (x:xs) (ys) = if (x `notElem` ys) then x:(filtrado xs ys)
                                                   else filtrado xs ys

bucleMientrasCambie::AFD->[(Int, Int)]->[(Int, Int)]
bucleMientrasCambie afd xs = if (xs /= nuevoXS) then bucleMientrasCambie afd nuevoXS
                                                else xs
  where nuevoXS = buclePrimerEstado afd xs 1

buclePrimerEstado::AFD->[(Int, Int)]->Int->[(Int, Int)]
buclePrimerEstado afd@(a, b, c, d, e) marcados qAct = if (qAct > b) then marcados
                                                                    else buclePrimerEstado afd (bucleSegundoEstado afd marcados qAct (qAct + 1)) (qAct + 1)

bucleSegundoEstado::AFD->[(Int, Int)]->Int->Int->[(Int, Int)]
bucleSegundoEstado afd@(a, b, c, d, e) marcados q1 qAct = if (qAct > b) then marcados
                                                                        else bucleSegundoEstado afd (bucleCaracter afd marcados q1 qAct a) q1 (qAct + 1)

bucleCaracter::AFD->[(Int, Int)]->Int->Int->[Char]->[(Int, Int)]
bucleCaracter afd marcados q1 q2 [] = marcados
bucleCaracter afd@(a, b, c, d, e) marcados q1 q2 (sigma:sigmas) = if ((sonEstadosDistinguibles marcados minQ maxQ) && ((q1, q2) `notElem` marcados)) then (q1, q2):marcados
                                                                                                                                                     else bucleCaracter afd marcados q1 q2 sigmas
  where maxQ = max (e ! (q1, sigma)) (e !(q2, sigma))
        minQ = min (e ! (q1, sigma)) (e !(q2, sigma))

{-sonEstadosDistinguibles::AFD->[(Int, Int)]->Int->Int->Bool
sonEstadosDistinguibles ([], b, c, d, e) marcados q1 q2 = False
sonEstadosDistinguibles (a:as, b, c, d, e) marcados q1 q2 = (tupla (e ! (q1, a)) (e ! (q2, a))) `elem` marcados || (sonEstadosDistinguibles (as, b, c, d, e) marcados q1 q2)
  where tupla x y = (x, y)-}
  
sonEstadosDistinguibles::[(Int, Int)]->Int->Int->Bool
sonEstadosDistinguibles marcados q1 q2 = (q1, q2) `elem` marcados

estadosDistinguiblesTriviales::AFD->[(Int, Int)]
estadosDistinguiblesTriviales (a, b, c, [], e) = []
estadosDistinguiblesTriviales afd@(a, b, c, (d:ds), e) = estadosDistinguiblesTriviales' afd 1 (d:ds)

estadosDistinguiblesTriviales'::AFD->Int->[Int]->[(Int, Int)]
estadosDistinguiblesTriviales' afd@(a, b, c, d, e) qActual [] = if (qActual < b) then estadosDistinguiblesTriviales' afd (qActual + 1) d
                                                                                 else []
estadosDistinguiblesTriviales' afd@(a, b, c, d, e) qActual (qFin:qFinales)
  | qActual < qFin = if (qActual `notElem` d) then (qActual, qFin):(estadosDistinguiblesTriviales' afd qActual qFinales)
                                              else estadosDistinguiblesTriviales' afd qActual qFinales
  | qActual > qFin = if (qActual `notElem` d) then (qFin, qActual):(estadosDistinguiblesTriviales' afd qActual qFinales)
                                              else estadosDistinguiblesTriviales' afd qActual qFinales
  | otherwise = estadosDistinguiblesTriviales' afd qActual qFinales

-- Construir afds complementarios y afds producto?

esEstadoFinal::AFD->Int->Bool
esEstadoFinal (a, b, c, d, e) q = elem q d

compartenAlfabeto::AFD->AFD->Bool
compartenAlfabeto (a, b, c, d, e) (a', b', c', d', e') = (aa == aa')
  where aa = ordenaSigma (eliminaRepetidos a)
        aa' = ordenaSigma (eliminaRepetidos a') 

sonIguales::AFD->AFD->Bool
sonIguales afd1 afd2 = (compartenAlfabeto afd1 afd2) && (not ((lenguajeInterseccionNoVacio afd'1 compAfd'2) || (lenguajeInterseccionNoVacio compAfd'1 afd'2)))
  where afd'1 = minimizaAutomata afd1
        afd'2 = minimizaAutomata afd2
        compAfd'1 = afdComplementario afd'1
        compAfd'2 = afdComplementario afd'2

afdComplementario::AFD->AFD
afdComplementario (a, b, c, d, e) = (a, b, c, (filter (`notElem` d) [1..b]), e)

lenguajeInterseccionNoVacio::AFD->AFD->Bool
lenguajeInterseccionNoVacio ([], b, c, d, e) afd2 = False
lenguajeInterseccionNoVacio afd1 ([], b, c, d, e) = False
lenguajeInterseccionNoVacio (a, b, c, [], e) afd2 = False
lenguajeInterseccionNoVacio afd1 (a, b, c, [], e) = False
lenguajeInterseccionNoVacio afd1@(a, b, c, d, e) afd2@(a', b', c', d', e') = recorreAFDproducto afd1 c afd2 c' a []

recorreAFDproducto::AFD->Int->AFD->Int->[Char]->[(Int, Int)]->Bool
recorreAFDproducto afd1 q1 afd2 q2 [] qVisitados = (esEstadoFinal afd1 q1) && (esEstadoFinal afd2 q2)
recorreAFDproducto afd1@(a, b, c, d, e) q1 afd2 q2 (s:sigmaAprobar) qVisitados = if (notElem (q1, q2) qVisitados) then (((esEstadoFinal afd1 q1) && (esEstadoFinal afd2 q2)) || 
                                                                                                                        recorreAFDproducto afd1 (ejecutaAFD afd1 q1 (s:[])) afd2 (ejecutaAFD afd2 q2 (s:[])) a ((q1, q2):qVisitados) ||
                                                                                                                        recorreAFDproducto afd1 q1 afd2 q2 sigmaAprobar qVisitados)
                                                                                                                  else False