
{-------------------------------------------------------------------------------
 -                                                                             -
 - Simulador y simplificador rudimientario de Automatas Finitos Deterministas  -
 -                                                                             - 
 - Creado por: Alejandro Dobeson Beirao                                        -
 - Contacto: granmacco@gmail.com ; www.sandoreus.net                           -
 -                                                                             -
 - Este software se distribuye "tal cual" bajo licencia aGPL v3.0              -
 - Compruebe los terminos dicha licencia en el archivo LICENSE.TXT incluido    -
 -                                                                             -
 -------------------------------------------------------------------------------}


module AFDmenus (menu, menuIni, cargarAfdVacio, opcionAfdDisco) where 

import AFDdata
import Char
import Array
import System.IO

menu:: AFD->IO ()             
menu afd@(a, b, c, d, e) = do putStrLn "Programa de simulacin de AFDs"
                              putStrLn "=============================="
                              putStrLn ""
                              escribeAFD afd
                              putStrLn ""
                              putStrLn "1. Calcular el AFD mnimo"
                              putStrLn "2. Ejecutar el AFD"
                              putStrLn "3. Aadir o eliminar caracteres a sigma"
                              putStrLn "4. Aadir o eliminar estados"
                              putStrLn "5. Modificar transiciones delta"
                              putStrLn "6. Cambiar el estado inicial"
                              putStrLn "7. Aadir o eliminar estados finales"
                              putStrLn "8. Comparar el AFD con otro dado"
                              putStrLn "9. Guardar el AFD actual en disco"
                              putStrLn "10. Cargar un nuevo AFD desde el disco"
                              putStrLn ""
                              putStrLn "0. Salir de la aplicacin"
                              putStrLn ""
                              opcion <- leeNat
                              putStrLn ""
                              case opcion of
                                  1 -> do afd' <- opcionMinimizaAFD afd
                                          menu afd'
                                  2 -> do opcionEjecutaAFD afd
                                          menu afd
                                  3 -> do afd' <- opcionAnadeOEliminaSigma afd
                                          menu afd'
                                  4 -> do afd' <- opcionAnadeOEliminaEstados afd 
                                          menu afd'
                                  5 -> do afd' <- opcionModificaDelta afd 
                                          menu afd'
                                  6 -> do afd' <- opcionModificaQi afd
                                          menu afd'
                                  7 -> do afd' <- opcionModificaQf afd
                                          menu afd'
                                  8 -> do opcionComparaAutomatas afd
                                          menu afd
                                  9 -> do opcionGuardaAFD afd
                                          menu afd
                                  10 -> do afd' <- opcionAfdDisco
                                           menu afd'
                                  0 -> do opcionSalir afd
                                  _ -> do menuError 
                                          menu afd

menuError::IO ()
menuError = do putStr "Opcin no reconocida. Escriba el nmero de la opcin para continuar."
               x <- getLine
               return ()

leeNat :: IO Int
leeNat = do cad <- getLine
            if ((all isDigit cad) && not (null cad)) then return (read cad)
                                                     else do putStrLn ""
                                                             putStrLn "Error, opcin no reconocida. Ha escrito lo que deba?"
                                                             putStrLn ""
                                                             leeNat

creaArrayVacio::Delta
creaArrayVacio = array ((0, 'b'), (0, 'a')) []

creaArray::(Int, Char)->(Int, Char)->[((Int,Char),Int)]->Delta
creaArray a b assoc = array (a, b) assoc

opcionAfdDisco::IO AFD
opcionAfdDisco = do putStrLn "Introduzca el nombre del fichero a cargar:"
                    fich <- getLine
                    afd <- cargarAfdDisco fich
                    return afd

cargarAfdVacio::AFD
cargarAfdVacio = ([], 1, 1, [], creaArrayVacio)

cargarAfdDisco::String->IO AFD
cargarAfdDisco fich = do lineas <- readFile fich
                         let (linea1:linea2:linea3:linea4:linea5) = lines lineas
                         let sigma = ordenaSigma linea1
                         let q = read linea2 :: Int
                         let qIni = read linea3 :: Int
                         let qFin = read linea4 :: [Int]
                         let delta = (creaArray (1, (head sigma)) ((read linea2 :: Int), (last sigma)) (leerDelta linea5))
                         if (null sigma) then return (sigma, q, qIni, qFin, creaArrayVacio)
                                         else return (sigma, q, qIni, qFin, delta)

leerDelta::[String]->[((Int, Char), Int)]
leerDelta [] = []
leerDelta (xs:xss)= (procesaDelta xs):(leerDelta xss)
  where procesaDelta s = ((qIni, ch), qFin)
          where qIni = read (takeWhile (/= ',') (drop 1 s)) :: Int
                ch = head (take 1 (drop 1 (dropWhile (/=',') s)))
                qFin = read (takeWhile (/= ')') (drop 3 (dropWhile (/=',') s))) :: Int

menuIni::IO Int
menuIni = do putStrLn "Programa de simulacin de AFDs"
             putStrLn "=============================="
             putStrLn ""
             putStrLn "1. Iniciar el programa con un AFD vaco"
             putStrLn "2. Iniciar el programa con un AFD predefinido (Fichero)"
             putStrLn ""
             opcion <- leeNat
             putStrLn ""
             if ((opcion > 0) && (opcion < 3)) then return opcion
                                               else do menuError
                                                       menuIni

menuAnade::String->IO (Int)
menuAnade x = do putStrLn ("1. Aadir un" ++ x)
                 putStrLn ("2. Eliminar un" ++ x)
                 opcion <- leeNat
                 if ((opcion > 0) && (opcion < 3)) then return opcion
                                                   else do menuError
                                                           menuAnade x

escribeAFD::AFD->IO()
escribeAFD (a, b, c, d, e) = do putStr "Sigma: "
                                escribeSigma a
                                putStr "Q: "
                                putStr (show b)
                                putStr ", Q inicial: "
                                putStr (show c)
                                putStr ", Q finales: "
                                putStrLn (show d)
                                putStr "Delta: "
                                putStrLn (show (assocs e))

escribeSigma::[Char]->IO()
escribeSigma [] = putStrLn "[]"
escribeSigma (x:xs) = do putStr ("[" ++ show x)
                         escribeSigma' (xs)
                         putStrLn "]"
  where escribeSigma' [] = return ()
        escribeSigma' (x:xs) = do putStr (", " ++ show x) 
                                  escribeSigma' xs

escribeDelta::Handle->[((Int, Char), Int)]->IO ()
escribeDelta outh [] = return ()
escribeDelta outh (x:xs) = do let q1 = (fst.fst) x
                              let ch = (snd.fst) x
                              let q2 = snd x
                              hPutStrLn outh ("(" ++ (show q1) ++ "," ++ (drop 1 (take 2 (show ch))) ++ "," ++ (show q2) ++ ")")
                              escribeDelta outh xs

opcionAnadeChar::AFD->Int->IO AFD
opcionAnadeChar afd 0 = return afd
opcionAnadeChar (a, b, c, d, e) n = if (null a) then do let ch = 'a'
                                                        afd' <- (opcionAnadeChar' (ordenaSigma (ch:a), b, c, d, e) 1 ch [])
                                                        opcionAnadeChar afd' (n-1)
                                                else do let ch = succ (last a)
                                                        afd' <- (opcionAnadeChar' (ordenaSigma (ch:a), b, c, d, e) 1 ch [])
                                                        opcionAnadeChar afd' (n-1)

opcionAnadeChar'::AFD->Int->Char->[((Int, Char), Int)]->IO AFD
opcionAnadeChar' afd@(a, b, c, d, e) q ch result =  if (q > b) then return (anadeDelta afd (result ++ oldAssocs))
                                                               else do putStrLn ("Est usted aadiendo la transicin delta que parte del estado q=" ++ show q ++ " con la etiqueta " ++ [ch])
                                                                       putStrLn ("A qu estado debera llevar la transicin (" ++ show q ++ ", " ++ [ch] ++ ")?")
                                                                       qf <- leeNat
                                                                       if (not (inBounds b qf)) then do putStrLn ("Error: Escriba un estado existente, es decir, comprendido entre 1 y " ++ show b ++ ".")
                                                                                                        linea <- getLine
                                                                                                        opcionAnadeChar' afd q ch result
                                                                                                else do putStrLn ("Transicin aadida entre el estado " ++ show q ++ " y el estado " ++ show qf)
                                                                                                        putStrLn ""
                                                                                                        opcionAnadeChar' afd (q+1) ch (((q, ch), qf):result)
  where oldAssocs = assocs e

opcionEliminaChar::AFD->Int->IO AFD
opcionEliminaChar afd 0 = return afd
opcionEliminaChar afd@([], b, c, d, e) n = return afd
opcionEliminaChar afd@(a, b, c, d, e) n = opcionEliminaChar (eliminaChar afd (last a)) (n-1)

opcionAnadeEstado::AFD->Int->IO AFD  
opcionAnadeEstado afd@(a, b, c, d, e) n = if (n > 0) then do putStrLn ("Est usted aadiendo el estado q=" ++ show (b+1))
                                                             afd' <- (opcionAnadeEstado' afd a [])
                                                             opcionAnadeEstado afd' (n-1)
                                                     else do return afd
  
opcionAnadeEstado'::AFD->[Char]->[((Int, Char), Int)]->IO AFD
opcionAnadeEstado' afd@(a, b, c, d, e) [] result = return (anadeDelta (a, (b+1), c, d, e) (result ++ oldAssocs))
  where oldAssocs = assocs e
opcionAnadeEstado' afd@(a, b, c, d, e) (x:xs) result = do let q = (b + 1)
                                                          putStrLn ("A que estado debera llevar la transicin (" ++ show q ++ ", " ++ show x ++ ")?")
                                                          qf <- leeNat
                                                          if (not (inBounds (b+1) qf)) then do putStrLn ("Error: Escriba un estado existente, es decir, comprendido entre 1 y " ++ show (b+1) ++ ".")
                                                                                               linea <- getLine
                                                                                               opcionAnadeEstado' afd (x:xs) result
                                                                                       else do let tupla = ((q, x), qf)
                                                                                               opcionAnadeEstado' afd xs (tupla:result)

opcionEliminaEstado::AFD->Int->IO AFD
opcionEliminaEstado afd@(a, b, c, d, e) q = if (hayArcosHaciaQ afd q) then do putStrLn "Error: Existen arcos que apuntan al estado que pretendes eliminar, en concreto:"
                                                                              putStrLn (show (arcosHaciaQ afd q))
                                                                              putStrLn "Tan solo se pueden eliminar estados hacia los que no apunte ningn arco. Modifiquelos antes de intentar eliminar el estado."
                                                                              linea <- getLine
                                                                              return afd
                                                                      else if (q == c) then do putStrLn "Error: Se est intentando eliminar el estado inicial. Si quiere eliminarlo, cambie el estado inicial a otro estado"
                                                                                               linea <- getLine
                                                                                               return afd
                                                                                       else return (eliminaEstado (a, b, c, d, e) q)
opcionEjecutaAFD::AFD->IO ()
opcionEjecutaAFD afd@(a, b, c, d, e) = do putStrLn "Ejecucin del AFD"
                                          putStrLn "================="
                                          putStrLn ""
                                          opcionEjecutaAFD' afd c []

opcionEjecutaAFD'::AFD->Int->[Char]->IO ()
opcionEjecutaAFD' afd@(a, b, c, d, e) qActual entradaAcum = do putStrLn ("Q actual=" ++ show qActual ++ ". Estado final=" ++ show (not (null (filter(==qActual) d))))
                                                               putStrLn ("Entrada evaluada hasta el momento: " ++ entradaAcum)
                                                               putStrLn "Introduzca la siguiente entrada a evaluar. Para salir, no escriba nada y pulse intro."
                                                               entrada <- getLine
                                                               if (null entrada) then return ()
                                                                                 else if (inSigma a entrada) then do opcionEjecutaAFD' afd (ejecutaAFD afd qActual entrada) (entradaAcum ++ entrada)
                                                                                                             else do putStrLn "Error: El carcter introducido no existe en sigma. Se ha equivocado al escribir?"
                                                                                                                     linea <- getLine
                                                                                                                     opcionEjecutaAFD' afd qActual entradaAcum

opcionMinimizaAFD::AFD->IO AFD
opcionMinimizaAFD afd@(a, b, c, d, e) = if (null a) then do putStrLn "Minimizando el AFD..."
                                                            putStrLn "Completado. Pulse INTRO para continuar..."
                                                            linea <- getLine
                                                            return cargarAfdVacio
                                                    else do putStrLn "Minimizando el AFD..."
                                                            let afd' = (minimizaAutomata afd)
                                                            putStrLn "Completado. Pulse INTRO para continuar..."
                                                            linea <- getLine
                                                            return afd'
                           
opcionAnadeOEliminaSigma::AFD->IO AFD
opcionAnadeOEliminaSigma afd@(a, b, c, d, e) = do opa <- menuAnade " caracter a sigma"
                                                  if (opa == 1) then do putStrLn "Cuntos caracteres quiere aadir a sigma?"
                                                                        num <- leeNat
                                                                        opcionAnadeChar afd num
                                                                else do putStrLn "Cuntos caracteres quiere eliminar de sigma?"
                                                                        num <- leeNat
                                                                        if (num <= (length a)) then opcionEliminaChar afd num
                                                                                               else do putStrLn ("Error: Sigma est compuesto por " ++ show (length a) ++ " caracteres. Introduzca un nmero menor o igual a la longitud de sigma.")
                                                                                                       return afd

---- Antigua opcion Aadir o Eliminar de sigma
{-                                                  if (opa == 1) then do putStrLn "Que carcter quiere aadir a Sigma?"
                                                                        ch <- getLine
                                                                        if ((length ch) > 1) then do putStrLn "Error: No se pueden aadir simbolos compuestos por ms de un caracter."
                                                                                                     linea <- getLine
                                                                                                     return afd
                                                                                             else if (null (filter(==(head ch)) a)) then do afd' <- opcionAnadeChar afd (head ch)
                                                                                                                                            putStrLn ("El carcter " ++ ch ++ " se aadi correctamente.")
                                                                                                                                            linea <- getLine
                                                                                                                                            return afd'
                                                                                                                                    else do putStrLn ("El carcter " ++ ch ++ " ya exista en sigma.")
                                                                                                                                            linea <- getLine
                                                                                                                                            return afd
                                                                else do putStrLn "Que carcter quiere eliminar de Sigma?"
                                                                        ch <- getLine
                                                                        if ((length ch) > 1) then do putStrLn "Error: No se pueden eliminar simbolos compuestos por ms de un caracter."
                                                                                                     linea <- getLine
                                                                                                     return afd
                                                                                             else if (not (null (filter(==(head ch)) a))) then do putStrLn ("El carcter " ++ ch ++ " se elimin correctamente.")
                                                                                                                                                  linea <- getLine
                                                                                                                                                  return (eliminaChar afd (head ch))
                                                                                                                                          else do putStrLn ("El carcter " ++ ch ++ " no exista en sigma.")
                                                                                                                                                  linea <- getLine
                                                                                                                                                  return afd-}

opcionAnadeOEliminaEstados::AFD->IO AFD
opcionAnadeOEliminaEstados afd@(a, b, c, d, e) = do opa <- menuAnade "os cuantos estado/s"
                                                    if (opa == 1) then do putStrLn "Cuantos estados nuevos quiere aadir?"
                                                                          num <- leeNat
                                                                          opcionAnadeEstado afd num
                                                                  else if (b <= 1) then do putStrLn "Error: No se pueden eliminar estados en un AFD vaco o con un slo estado. Aade algn estado antes de intentar eliminarlo."
                                                                                           linea <- getLine
                                                                                           return afd
                                                                                   else do putStrLn "Que estado quiere eliminar?"
                                                                                           num <- leeNat
                                                                                           if (num > b) then do putStrLn ("Error: No se puede eliminar un estado que no existe. Recuerda, los estados van desde 1 hasta " ++ show b)
                                                                                                                linea <- getLine
                                                                                                                return afd
                                                                                                        else do afd' <- opcionEliminaEstado afd num
                                                                                                                return afd'

opcionModificaDelta::AFD->IO AFD
opcionModificaDelta afd@(a, b, c, d, e) = do putStrLn "A qu estado pertenece el delta que quieres modificar?"
                                             qi <- leeNat
                                             if (not (inBounds b qi)) then do putStrLn ("Error: El estado " ++ show qi ++ " no existe.")
                                                                              linea <- getLine
                                                                              return afd
                                                                      else do putStrLn "Qu arco quieres modificar?"
                                                                              ch <- getLine
                                                                              if (null (filter (==(head ch)) a)) then do putStrLn ("Error: El caracter " ++ ch ++ " no se encuentra en sigma.")
                                                                                                                         linea <- getLine
                                                                                                                         return afd
                                                                                                                 else do putStrLn "A qu estado quieres que apunte ahora?"
                                                                                                                         qf <- leeNat
                                                                                                                         if (not (inBounds b qf)) then do putStrLn ("Error: El estado " ++ show qf ++ " no existe.")
                                                                                                                                                          linea <- getLine
                                                                                                                                                          return afd
                                                                                                                                                  else do putStrLn "Se ha modificado delta satisfactoriamente"
                                                                                                                                                          linea <- getLine
                                                                                                                                                          return (modificaDelta afd qi (head ch) qf)

opcionModificaQi::AFD->IO AFD
opcionModificaQi afd@(a, b, c, d, e) = do putStrLn "Que estado quiere que sea el nuevo estado inicial?"
                                          qi <- leeNat
                                          if (not (inBounds b qi)) then do putStrLn ("Error: El nuevo estado inicial " ++ show qi ++ " no existe.")
                                                                           linea <- getLine
                                                                           return afd
                                                                   else do putStrLn "Se ha modificado el estado inicial satisfactoriamente."
                                                                           linea <- getLine
                                                                           return (a, b, qi, d, e)

opcionModificaQf::AFD->IO AFD
opcionModificaQf afd@(a, b, c, d, e) = do opa <- menuAnade " estado final"
                                          if (opa == 1) then do putStrLn "Qu estado ser ahora final?"
                                                                num <- leeNat
                                                                if (not (inBounds b num)) then do putStrLn "Error: el estado indicado no existe."
                                                                                                  linea <- getLine
                                                                                                  return afd
                                                                                          else if (not (null (filter (==num) d))) then do putStrLn "Error: El estado indicado ya habia sido aadido anteriormente como estado final."
                                                                                                                                          linea <- getLine
                                                                                                                                          return afd
                                                                                                                                  else do putStrLn "Estado final aadido satisfactoriamente."
                                                                                                                                          linea <- getLine
                                                                                                                                          return (a, b, c, num:d, e)
                                                        else if (b <= 0) then do putStrLn "Error: El AFD actual no tiene estados finales."
                                                                                 linea <- getLine
                                                                                 return afd
                                                                         else do putStrLn "Que estado final quiere eliminar?"
                                                                                 num <- leeNat
                                                                                 if (null (filter (==num) d)) then do putStrLn "Error: El estado indicado no es final"
                                                                                                                      linea <- getLine
                                                                                                                      return afd
                                                                                                              else do putStrLn "Estado final eliminado satisfactoriamente."
                                                                                                                      linea <- getLine
                                                                                                                      return (a, b, c, filter (/=num) d, e)                                                                           
                                                                           
opcionSalir::AFD->IO ()
opcionSalir afd = do putStrLn "Desea guardar el autmata actual antes de salir? (S/N)"
                     linea <- getLine
                     let ch = (head linea)
                     if ((not (null linea)) && ((ch == 's') || (ch == 'S'))) then opcionGuardaAFD afd
                                                                             else return ()
                     putStrLn "Terminando programa..."
                     
opcionGuardaAFD::AFD->IO()
opcionGuardaAFD afd@(a, b, c, d, e) = do putStrLn "Introduzca el nombre del fichero en el que guardar el AFD. Se aadir automticamente la extensin '.txt':"
                                         fich <- getLine
                                         outh <- openFile (fich ++ ".txt") WriteMode
                                         hPutStrLn outh a
                                         hPutStrLn outh (show b)
                                         hPutStrLn outh (show c)
                                         hPutStrLn outh (show d)
                                         escribeDelta outh (assocs e)
                                         hClose outh
                                         putStrLn ("AFD guardado en " ++ fich ++ ".txt")
                                         linea <- getLine
                                         return ()

opcionComparaAutomatas::AFD->IO ()
opcionComparaAutomatas afd = do putStrLn "A continuacin se va a proceder a cargar el autmata con el que comparar"
                                afd' <- opcionAfdDisco
                                if (sonIguales afd afd') then putStrLn "Ambos automatas SI expresan el mismo lenguaje"
                                                         else putStrLn "Ambos automatas NO expresan el mismo lenguaje"
                                ch <- getLine
                                return ()